#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "dafst.h"
#include "probtool.h"
#include "bin2txt.h"
#include "load_double.h"

#ifdef WIN32
#undef USING_MMAP
#endif

#ifdef USING_MMAP
#include <sys/mman.h> // mmap
#include <sys/stat.h> // open
#include <fcntl.h> // open
#include <sys/types.h> // open
#include <unistd.h> // close
#endif

////////////////////////////////////////////////////////////////////////////////
// ޸ 
void prob_CloseFST
(
	PROBtFST *ProbFst
) 
{
	if (! ProbFst) return;

	// fst
	if (ProbFst->fst) fst_Close( ProbFst->fst);
	
	// information
	if (ProbFst->info)
	{
		b2t_Close( ProbFst->info);
	}

#ifdef USING_MMAP
	// munmap
	if (ProbFst->mmapsize > 0)
	{
		if ( munmap( (void *)ProbFst->prob, ProbFst->mmapsize) == -1)
		{
			fprintf(stderr, "munmap error\n");
		}
	}
#else
	// ޸ 
	if (ProbFst->prob)
	{
		free( ProbFst->prob);
		ProbFst->prob = NULL;
	}
#endif

	free( ProbFst);
	ProbFst = NULL;
}

////////////////////////////////////////////////////////////////////////////////
// fst ϰ Ȯ  оδ.
PROBtFST *prob_OpenFST
(
	const char *FstFileName, 
	const char *ProbFileName
)
{
	PROBtFST *prob_fst = (PROBtFST *)malloc( sizeof(PROBtFST));
	assert (prob_fst != NULL);
	
	// ʱȭ
	prob_fst->mmapsize = -1;
	prob_fst->info = NULL;
	prob_fst->fst = NULL;
	prob_fst->prob = NULL;
	
	// fst 
	fprintf( stderr, "\tLoading FST file.. [%s]", FstFileName);
	if ( (prob_fst->fst = fst_Open( FstFileName, NULL)) == NULL)
	{
		fprintf( stderr, "Load failure! [%s]\n", FstFileName);
		prob_CloseFST( prob_fst);
		return NULL;
	}
	fprintf( stderr, "\t[done]\n");
	
	// Ȯ ȭ 
	fprintf( stderr, "\tLoading probability file.. [%s]", ProbFileName);
#ifdef USING_MMAP
	if ((prob_fst->prob = prob_LoadFloat( ProbFileName, &(prob_fst->mmapsize))) == NULL) // Ȯ б
#else
	if ((prob_fst->prob = prob_LoadFloat( ProbFileName)) == NULL) // Ȯ б
#endif
	{
		fprintf( stderr, "Load failure! [%s]\n", ProbFileName);
		prob_CloseFST( prob_fst);
		return NULL;
	}
	fprintf( stderr, "\t[done]\n");

	return prob_fst;
}

////////////////////////////////////////////////////////////////////////////////
// fst ϰ Ȯ  оδ.
PROBtFST *prob_OpenWithInfoFST
(
	const char *FstFileName, // и  FST 
	const char *InfoFileName, //  Ʈ 
	const char *ProbFileName // Ȯ 
) 
{
	PROBtFST *prob_fst = (PROBtFST *)malloc( sizeof(PROBtFST));
	assert (prob_fst != NULL);

	// ʱȭ
	prob_fst->mmapsize = -1;
	prob_fst->info = NULL;
	prob_fst->fst = NULL;
	prob_fst->prob = NULL;

	// fst 
	fprintf( stderr, "\tLoading FST file.. [%s]", FstFileName);
	if ( (prob_fst->fst = fst_Open( FstFileName, NULL)) == NULL)
	{
		fprintf( stderr, "Load failure [%s]\n", FstFileName);
		prob_CloseFST( prob_fst);
		return NULL;
	}
	fprintf( stderr, "\t[done]\n");
	
	//  ȭ 
	fprintf( stderr, "\tLoading information file.. [%s]", InfoFileName);

	prob_fst->info = b2t_Open( InfoFileName);
	
	if (prob_fst->info == NULL)
	{
		fprintf( stderr, "b2t_Open error [%s]\n", InfoFileName);
		prob_CloseFST( prob_fst);
		return NULL;
	}
	fprintf(stderr, "\t[done]\n");

	// Ȯ ȭ 
	fprintf(stderr, "\tLoading probability file.. [%s]", ProbFileName);

#ifdef USING_MMAP
	if ((prob_fst->prob = prob_LoadFloat( ProbFileName, &(prob_fst->mmapsize))) == NULL) // Ȯ б
#else
	if ((prob_fst->prob = prob_LoadFloat( ProbFileName)) == NULL) // Ȯ б
#endif
	{
		fprintf( stderr, "Load failure! [%s]\n", ProbFileName);
		prob_CloseFST( prob_fst);
		return NULL;
	}
	fprintf( stderr, "\t[done]\n");

	return prob_fst;
}
