#ifndef __probtool_H__
#define __probtool_H__

#ifdef WIN32
#undef USING_MMAP
#else
#define USING_MMAP // mmap
#endif

#ifdef __cplusplus
extern "C" {
#endif

typedef struct
{
  void	*fst;
  void	*info; // 사용하지 않으면 -1
  float	*prob;
  int	mmapsize; // mmap을 사용하지 않으면 -1
} PROBtFST;

////////////////////////////////////////////////////////////////////////////////
#define PROBcALMOST_ZERO	1.0e-40 // -92 /* 완전히 0을 쓰게 되면 log(0) = inf가 되어서 문제 발생 가능 */
                                        // 1.0e-100 // -230
#define PROBcLOG_ALMOST_ZERO	-100.0

#define PROBcNULLSTRING		"***"

#define PROBcDELIM_INTER	0x05 // 5   // 확률에서 분자와 분모의 경계 구분을 위함 예) P(D|A,B,C)= f(A,B,C,D)/f(A,B,C)에서 D와 A,B,C 간의 구분을 위한 것
#define PROBcDELIM_INTRA	0x07 // 128 // 문자열 내부의 경계 구분을 위함 예) P(D|A,B,C)= f(A,B,C,D)/f(A,B,C)에서 A,B,C 서로 간의 구분을 위한 것

////////////////////////////////////////////////////////////////////////////////
extern PROBtFST *prob_OpenFST
(
  const char *FstFileName, 
  const char *ProbFileName
);

extern PROBtFST *prob_OpenWithInfoFST
(
  const char *FstFileName,
  const char *InfoFileName,
  const char *ProbFileName
);
    
extern void prob_CloseFST
(
  PROBtFST *ProbFst
);
                        
/////
extern double prob_GetFSTProb1
(
  PROBtFST *ProbFst,
  char *Target
);

extern double prob_GetFSTProb2
(
  PROBtFST *ProbFst,
  char *Target, 
  char *T4
);

extern double prob_GetFSTProb3
(
  PROBtFST *ProbFst,
  char *Target, 
  char *T3, 
  char *T4
);

extern double prob_GetFSTProb4
(
  PROBtFST *ProbFst,
  char *Target, 
  char *T2, 
  char *T3, 
  char *T4
);

extern double prob_GetFSTProb5
(
  PROBtFST *ProbFst,
  char *Target, 
  char *T1, 
  char *T2, 
  char *T3,
  char *T4
);

extern double prob_GetFSTProb1a
(
  PROBtFST *prob_fst[],
  char *num
);

extern double prob_GetFSTProb2a
(
  PROBtFST *prob_fst[],
  char *num, 
  char *t4
);

extern double prob_GetFSTProb3a
(
  PROBtFST *prob_fst[],
  char *num, 
  char *t3, 
  char *t4
);

extern double prob_GetFSTProb4a
(
  PROBtFST *prob_fst[],
  char *num, 
  char *t2, 
  char *t3, 
  char *t4
);

extern double prob_GetFSTProb5a
(
  PROBtFST *prob_fst[],
  char *num, 
  char *t1, 
  char *t2, 
  char *t3,
  char *t4
);

extern double prob_GetFSTProbMLE
(
  PROBtFST *ProbFst,
  char *Left,
  char *Right
);

#ifdef __cplusplus
}
#endif // __cplusplus
              
#endif
